from datetime import datetime
from functools import wraps
from marshmallow import post_dump, pre_dump
from wedeliver_core_plus import WedeliverCorePlus
from wedeliver_core_plus.app_decorators import (
    handle_response,
    handle_auth,
    handle_exceptions,
    serializer,
)
from wedeliver_core_plus.helpers.time import DATETIME_FORMAT_SCHEMA

# from wedeliver_core_plus.helpers.get_prefix import get_prefix

# Dictionary to store route metadata
route_metadata = {}


def route(path, methods=["GET"], schema=None, many=False, allowed_roles=None, require_auth=True,
          append_auth_args=None,
          pre_login=False,
          allowed_permissions=None, guards=[]):
    app = WedeliverCorePlus.get_app()

    def factory(func):
        # Store metadata
        route_metadata[func.__name__] = {
            'module_path': getattr(func, 'testing_class_module', None),
            'url': path,
            'methods': methods,
            'schema': schema,
            'allowed_roles': allowed_roles,
            'is_testable': getattr(func, 'is_testable', False),
        }

        @app.route(path, methods=methods)
        @handle_response
        @handle_exceptions
        @handle_auth(require_auth=require_auth, append_auth_args=append_auth_args, allowed_roles=allowed_roles,
                     allowed_permissions=allowed_permissions, pre_login=pre_login, guards=guards)
        @serializer(schema=schema, many=many)
        @wraps(func)
        def decorator(*args, **kwargs):
            return func(*args, **kwargs)

        return decorator

    return factory


def testing_class(test_cls):
    def decorator(func):
        func.testing_class_module = getattr(test_cls, 'module_path', None)
        func.is_testable = True
        # Set the route function name on the test class
        setattr(test_cls, 'route_function_name', func.__name__)
        return func
    return decorator


def advance(features_class):
    def decorator(cls):
        original_cls = cls  # Save a reference to the original class

        # class_name = f"AutoAdvanced_{cls.__name__}"

        # Dynamically create the new class
        # AutoAdvancedClass = type(class_name, (cls, features_class), {})

        # Alternatively, you can use a regular class definition, if you prefer:
        # class AutoAdvancedClass(cls, features_class):
        #     pass

        class AutoAdvancedClass(cls, features_class):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def before_insert(self, mapper, connection):
                # Check if features_class has the method before_insert
                if hasattr(features_class, 'before_insert'):
                    features_class.before_insert(self, mapper, connection)

            def after_insert(self, mapper, connection):
                # Check if features_class has the method before_insert
                if hasattr(features_class, 'after_insert'):
                    features_class.after_insert(self, mapper, connection)

            def before_update(self, mapper, connection):
                if hasattr(features_class, 'before_update'):
                    features_class.before_update(self, mapper, connection)

            def after_update(self, mapper, connection):
                if hasattr(features_class, 'after_update'):
                    features_class.after_update(self, mapper, connection)

            def before_delete(self, mapper, connection):
                if hasattr(features_class, 'before_delete'):
                    features_class.before_delete(self, mapper, connection)

            def after_delete(self, mapper, connection):
                if hasattr(features_class, 'after_delete'):
                    features_class.after_delete(self, mapper, connection)

        # Pass the original class to restfull
        return restfull(original_cls, AutoAdvancedClass)

    return decorator


def restfull(original_cls, advanced_model_class):
    # from flask import current_app

    model_name = original_cls.__name__.lower()
    app = WedeliverCorePlus.get_app()
    db = app.extensions['sqlalchemy'].db
    ma = app.extensions['flask-marshmallow']

    module_path = original_cls.__module__

    # Split the string by the dot character
    parts = module_path.split('.')

    # Extract the third part of the module path
    service_name = f'/{parts[2]}' if len(parts) > 2 else None

    single_name = model_name if not hasattr(advanced_model_class,
                                            '__restfull__') else advanced_model_class.__restfull__.get('model_name',
                                                                                                       model_name)
    plural_name = f'{model_name}s' if not hasattr(advanced_model_class,
                                                  '__restfull__') else advanced_model_class.__restfull__.get(
        'models_name', f'{model_name}s')

    api_prefix = '/api/v1' if not hasattr(advanced_model_class,
                                          '__restfull__') else advanced_model_class.__restfull__.get('url_prefix',
                                                                                                     '/api/v1')

    api_prefix = f'{service_name}{api_prefix}' if service_name else api_prefix

    def generate_dynamic_schema(method):
        required_fields = [] if not hasattr(original_cls, '__restfull__') else original_cls.__restfull__.get('methods',
                                                                                                             {}).get(
            method, {}).get('required_fields', [])

        hidden_fields = [] if not hasattr(original_cls, '__restfull__') else original_cls.__restfull__.get(
            'hidden_fields', [])

        class ModelSchema(ma.SQLAlchemyAutoSchema):
            class Meta:
                # model = original_cls
                table = original_cls.__table__
                # include_relationships = True
                # load_instance = True

            @post_dump
            def format_datetime_fields(self, data, **kwargs):
                for key, value in data.items():
                    # Check if the value is a datetime string by trying to convert it
                    try:
                        dt_obj = datetime.fromisoformat(value.replace("T", " "))
                        # If successful, reformat it
                        data[key] = dt_obj.strftime(DATETIME_FORMAT_SCHEMA)
                    except (ValueError, TypeError, AttributeError):
                        # Skip values that can't be converted to datetime
                        continue
                return data

            @post_dump(pass_many=True)
            def filter_fields(self, data, many, **kwargs):
                if many:
                    return [{key: item[key] for key in item if key not in hidden_fields} for item in data]
                return {key: data[key] for key in data if key not in hidden_fields}

        for field_name, field_obj in ModelSchema._declared_fields.items():
            if field_name in required_fields:
                field_obj.required = True
            else:
                field_obj.required = False

        return ModelSchema

    get_model_schema = generate_dynamic_schema('get')
    list_model_schema = generate_dynamic_schema('list')
    post_model_schema = generate_dynamic_schema('post')

    # get_model_schema.Meta.model = original_cls
    @route(f'{api_prefix}/resource/{single_name}', methods=['POST'], schema=post_model_schema)
    def create(validated_data={}):
        new_item = advanced_model_class(**validated_data)
        db.session.add(new_item)
        db.session.commit()
        return new_item

    @route(f'{api_prefix}/resource/{single_name}/<int:id>', methods=['GET'], schema=get_model_schema)
    def read(id, validated_data={}):
        item = advanced_model_class.query.get_or_404(id)
        return item

    @route(f'{api_prefix}/resource/{plural_name}', methods=['GET'], schema=list_model_schema)
    def list(validated_data={}):
        items = advanced_model_class.query.all()
        return items

    @route(f'{api_prefix}/resource/{single_name}/<int:item_id>', methods=['PUT'])
    def update(item_id, validated_data={}):
        item = advanced_model_class.query.get_or_404(item_id)
        for key, value in validated_data.items():
            setattr(item, key, value)
        db.session.commit()
        return True

    @route(f'{api_prefix}/resource/{single_name}/<int:item_id>', methods=['DELETE'])
    def delete(item_id, validated_data={}):
        item = advanced_model_class.query.get_or_404(item_id)
        db.session.delete(item)
        db.session.commit()
        return True



import pytest


class BaseTestClass:
    """
    Base class for reusable test functionality.
    """

    module_path = None
    route_function_name = None

    @pytest.fixture(autouse=True)
    def base_setup(
            self,
            api_request,
            db_session,
            model_factory,
            web_auth_token,
            app_auth_token,
            app_auth_pre_token,
            auto_patch_microfetcher,
            request,
    ):
        # Inject common fixtures
        self.api_request = api_request
        self.db_session = db_session
        self.model_factory = model_factory
        self.web_auth_token = web_auth_token
        self.app_auth_token = app_auth_token
        self.app_auth_pre_token = app_auth_pre_token
        self.auto_patch_microfetcher = auto_patch_microfetcher

        # Allow derived classes to add additional fixtures
        self._custom_setup(request)

    def _custom_setup(self, request):
        """
        Method for derived classes to extend the setup process.
        """
        pass

    def patch_microfetcher(self, mock_microfetcher=None, module=None, **kwargs):
        """
        Patches the microfetcher with a given route and data. Automatically sets manual flag if the route differs from
        the default route_function_name.

        :param module: The route to patch. Defaults to self.route_function_name.
        :param mock_microfetcher: The MockMicroFetcher instance to use for patching.
        :param kwargs: Any additional keyword arguments to pass.
        """
        # Default to the class-level route_function_name if no route is provided
        if module is None:
            module = self.route_function_name

        # Automatically add manual flag if the route is different from the default
        if module != self.route_function_name:
            kwargs['manual'] = True

        # Call the actual microfetcher patching function
        self.auto_patch_microfetcher(module, mock_microfetcher, **kwargs)

    def call_api_request(self, **kwargs):

        return self.api_request(self.route_function_name, **kwargs)



from wedeliver_core_plus.helpers.topics import Topics
from wedeliver_core_plus.helpers import kafka_producers


def log_model_changes(
        data,
        **kwargs
):
    """
    Log model changes to kafka
    """
    kafka_producers.send_topic(topic=Topics().LOG_MODEL_CHANGES, datajson=dict(
        function_name='app.business_logic.logs.save_model_changes.execute',
        function_params=data
    ), kafka_configs=kwargs.get('kafka_configs'))



from wedeliver_core_plus.app_decorators import handle_exceptions


def raise_exception_in_background(error):
    @handle_exceptions
    def raise_error(e):
        raise e

    try:
        raise_error(error)
    except Exception:
        return


import flask_sqlalchemy
from flask import g
# from requests.adapters import HTTPAdapter
# from urllib3.util.retry import Retry

from wedeliver_core_plus import WedeliverCorePlus
from wedeliver_core_plus.helpers.exceptions import AppMicroFetcherError, AppFetchServiceDataError
import requests
import json

from wedeliver_core_plus.helpers.get_obj_value import get_obj_value, set_obj_value

from wedeliver_core_plus.helpers.service_config import ServiceConfig

from wedeliver_core_plus.helpers.enums import QueryTypes


class MicroFetcher(object):
    base_data = None
    app = None
    service_name = None
    service_url = None
    fields = None
    table_name = None
    column_name = None
    compair_operator = None
    column_values = None
    output_key = None
    lookup_key = None
    module_name = None
    function_params = None
    query_type = None
    search_list = None
    configs = None

    def __init__(self, service):
        self.app = WedeliverCorePlus.get_app()

        if isinstance(service, ServiceConfig):
            service = service.initialize()
            service_name = service.name
            service_url = service.url
        else:
            service_name = service
            service_url = None

        env_service_url = self.app.config.get(service_name)

        self.service_url = env_service_url if env_service_url else service_url

        self.service_name = service_name
        if not self.service_url:
            raise AppMicroFetcherError(f'Service {service_name} not defined on Env or in ServiceConfig')

    def join(self, base_data, output_key=None):
        self.base_data = base_data
        self.query_type = QueryTypes.SIMPLE_TABLE.value
        if output_key:
            output_key = output_key.split('as ')[1]

        self.output_key = "{}".format(self.service_name.split('_')[0].lower()) if not output_key else output_key
        return self

    def config(self, **configs):
        self.configs = configs
        return self

    def _prepare_search_list(self):
        output = dict()
        for index, item in enumerate(self.base_data):
            for search_column in self.search_configs.get("search_priority"):
                sanitize = None
                if isinstance(search_column, dict):
                    search_column_name = search_column.get('key')
                    operator = search_column.get('operator') or "IN"
                    sanitize = search_column.get('sanitize')
                else:
                    search_column_name = search_column
                    operator = 'IN'

                value = item.get(search_column_name)
                if sanitize and isinstance(sanitize, list):
                    for _san in sanitize:
                        value = _san(value)

                if value:
                    if not output.get(search_column_name):
                        output[search_column_name] = dict(
                            search_key=search_column_name,
                            operator=operator,
                            inputs=dict()
                        )
                    if not output[search_column_name]['inputs'].get(value):
                        output[search_column_name]['inputs'][value] = dict(
                            indexes=[index],
                            search_value=value
                        )
                    else:
                        output[search_column_name]['inputs'][value]["indexes"].append(index)
                    break

        output = list(output.values())
        for item in output:
            item['inputs'] = list(item['inputs'].values())

        self.search_list = output

    def search_config(self, configs):
        self.search_configs = configs
        self._prepare_search_list()
        return self

    def global_configs(self, **keywords):
        self.global_configs = keywords
        return self

    def feed_list(self, base_data, output_key=None):
        join_result = self.join(base_data, output_key)
        self.query_type = QueryTypes.SEARCH.value
        return join_result

    def select(self, *args):
        self.fields = list(args)
        return self

    def filter(self, *args):
        against = args[0].split('.')
        self.compair_operator = args[1]
        self.lookup_key = args[2]
        self.column_values = set()
        if isinstance(self.base_data, dict):
            self.column_values.add(get_obj_value(self.base_data, self.lookup_key))
        else:
            if isinstance(self.base_data, flask_sqlalchemy.Pagination):
                data = self.base_data.items
            else:
                data = self.base_data

            if isinstance(data, list):
                for row in data:
                    self.column_values.add(get_obj_value(row, self.lookup_key))
            else:
                self.column_values.add(get_obj_value(data, self.lookup_key))

        if not len(self.column_values):
            self.column_values = []
            return self
            # raise AppMicroFetcherError('Lookup key {} not found'.format(self.lookup_key))

        self.column_values = list(filter(None, self.column_values))
        self.column_values = self.column_values[0] if len(self.column_values) == 1 else self.column_values

        if self.compair_operator not in ('=', 'IN'):
            raise AppMicroFetcherError('Only == currently supported')

        self.compair_operator = 'IN' if isinstance(self.column_values, list) else self.compair_operator

        if len(against) != 2:
            self.column_name = against[0]
        else:
            self.table_name = against[0]
            self.column_name = against[1]

            self.fields.append(self.column_name)

        return self

    def fetch(self):
        if self.column_values or self.module_name or self.query_type == QueryTypes.SEARCH.value:
            return self._call_api()
        else:
            return self.base_data

    def execute(self):
        return self.fetch()

    def with_params(self, **kwargs):
        self.function_params = kwargs
        return self

    def from_function(self, module_name):
        self.query_type = QueryTypes.FUNCTION.value
        self.module_name = module_name
        return self

    def _call_api(self):

        url = "{}/fetch_relational_data".format(self.service_url)

        payload_dict = dict(
            __user_auth_data__=g.user if hasattr(g, 'user') else None,
        )
        if self.query_type in [QueryTypes.SIMPLE_TABLE.value, QueryTypes.FUNCTION.value]:
            payload_dict.update(
                dict(
                    fields=self.fields,
                    table_name=self.table_name,
                    column_name=self.column_name,
                    compair_operator=self.compair_operator,
                    column_values=self.column_values,
                )
            )

            if self.query_type == QueryTypes.FUNCTION.value:
                payload_dict.update(
                    functions=[
                        dict(
                            name=self.module_name,
                            fields=self.fields,
                            params=self.function_params if self.function_params else dict()
                        )
                    ]
                )

        elif self.query_type == QueryTypes.SEARCH.value:
            payload_dict.update(dict(
                query_type=self.query_type,
                table_name=self.search_configs.get("table_name"),
                search_list=self.search_list,
                append_extra=self.search_configs.get("append_extra"),
                use_country_code=self._get_global_config(key="use_country_code", default_value=True),
            ))

        payload = json.dumps(payload_dict)
        headers = {
            'country_code': self._get_global_config(key="country_code", default_value="sa"),
            'Content-Type': 'application/json',
        }

        session = requests  # requests.Session()

        # retry = Retry(
        #     total=5,
        #     backoff_factor=1,
        #     status_forcelist=[429, 500, 502, 503, 504],
        # )
        # adapter = HTTPAdapter(max_retries=retry)
        # session.mount('http://', adapter)
        # session.mount('https://', adapter)

        def _raise_error(message, error):
            self.app.logger.error(self.service_name)
            self.app.logger.error(message)
            self.app.logger.error(payload)
            self.app.logger.error(str(conn_err))
            raise AppFetchServiceDataError(message)

        response = None  # Initialize the response variable

        try:
            # response = requests.request("POST", url, headers=headers, data=payload)
            response = session.post(url, headers=headers, data=payload)
            # response.raise_for_status()
        except requests.exceptions.HTTPError as http_err:
            _raise_error('HTTP error occurred', http_err)
        except requests.exceptions.ConnectionError as conn_err:
            _raise_error('Connection error occurred', conn_err)
        except requests.exceptions.Timeout as timeout_err:
            _raise_error('Timeout error occurred', timeout_err)
        except requests.exceptions.RequestException as req_err:
            _raise_error('Request error occurred', req_err)
        except Exception as req_err:
            _raise_error('Exception occurred', req_err)

        if response is not None and response.status_code != 200:
            self.app.logger.error(self.service_name)
            self.app.logger.error(response.text)
            custom_text = None
            try:
                custom_text = json.loads(response.text).get('message')
            except json.decoder.JSONDecodeError:
                pass
            raise AppFetchServiceDataError(custom_text, response.status_code)

        result = response.json() if response else dict()

        if self.base_data is not None:
            return self._map_base(result)

        return result

    def _get_global_config(self, key, default_value):
        if hasattr(self, 'global_configs') and isinstance(self.global_configs, dict):
            if self.global_configs.get(key) is None:
                return default_value
            else:
                return self.global_configs.get(key)
        else:
            return default_value

    def _map_base(self, result):
        if self.query_type == QueryTypes.SEARCH.value:
            # map search result with the original object.
            _result = result.get("result")
            for item in _result:
                for _input in item.get('inputs'):
                    for _index in _input.get('indexes'):
                        self.base_data[_index][self.output_key] = _input.get('matched_id')
                        append_extra = self.search_configs.get('append_extra') if isinstance(self.search_configs,
                                                                                             dict) else []
                        for _ap_col in append_extra:
                            self.base_data[_index][_ap_col] = _input.get(_ap_col) if _input.get('matched_id') else \
                                self.base_data[_index].get(_ap_col)

            validation_result = []
            for _val in result.get("validation"):
                for _ind in _val.get("indexes"):
                    _val.pop("indexes", None)
                    validation_result.append(dict(
                        index=_ind,
                        **_val
                    ))
            return validation_result
        else:
            if isinstance(self.base_data, dict):
                if self.query_type == QueryTypes.SIMPLE_TABLE.value:
                    for rd in result:
                        if self.base_data.get(self.lookup_key) == rd.get(self.column_name):
                            try:
                                setattr(self.base_data, self.output_key, rd)
                            except AttributeError:
                                if self.base_data.get(self.output_key):
                                    if not isinstance(self.base_data[self.output_key], list):
                                        self.base_data[self.output_key] = [self.base_data[self.output_key]]

                                    self.base_data[self.output_key].append(
                                        rd
                                    )
                                else:
                                    self.base_data[self.output_key] = rd

                else:
                    for rd in result:
                        if self.base_data.get(self.lookup_key) == rd.get(self.column_name):
                            try:
                                setattr(self.base_data, self.output_key, rd)
                            except AttributeError:
                                self.base_data[self.output_key] = rd
            else:
                if isinstance(self.base_data, flask_sqlalchemy.Pagination):
                    data = self.base_data.items
                else:
                    data = self.base_data

                append_if_exists = self.configs.get("append_if_exists", False) if isinstance(self.configs,
                                                                                             dict) else False

                def _feed_base(_base, _data):
                    if isinstance(_data, list):
                        for _d in _data:
                            if str(get_obj_value(_base, self.lookup_key)) == str(_d.get(self.column_name)):
                                set_obj_value(_base, self.output_key, _d, append_if_exists)
                    else:
                        if str(get_obj_value(_base, self.lookup_key)) == _data.get(self.column_name):
                            set_obj_value(_base, self.output_key, _data, append_if_exists)

                if isinstance(data, list):
                    for row in data:
                        _feed_base(_base=row, _data=result)
                else:
                    _feed_base(_base=data, _data=result)

                if isinstance(self.base_data, flask_sqlalchemy.Pagination):
                    self.base_data.items = data
                else:
                    self.base_data = data

            return self.base_data



from unittest.mock import MagicMock
from enum import Enum


# Placeholder exception classes
class AppMicroFetcherError(Exception):
    pass


class AppFetchServiceDataError(Exception):
    pass


# Utility functions for getting and setting values in nested objects
def get_obj_value(obj, key):
    keys = key.split('.')
    for k in keys:
        if isinstance(obj, dict):
            obj = obj.get(k)
        else:
            obj = getattr(obj, k, None)
        if obj is None:
            return None
    return obj


def set_obj_value(obj, key, value, append_if_exists=False):
    keys = key.split('.')
    for k in keys[:-1]:
        if isinstance(obj, dict):
            obj = obj.setdefault(k, {})
        else:
            if not hasattr(obj, k):
                setattr(obj, k, {})
            obj = getattr(obj, k)
    last_key = keys[-1]

    if isinstance(obj, dict):
        if append_if_exists and last_key in obj:
            if not isinstance(obj[last_key], list):
                obj[last_key] = [obj[last_key]]
            obj[last_key].append(value)
        else:
            obj[last_key] = value
    else:
        if append_if_exists and hasattr(obj, last_key):
            existing_value = getattr(obj, last_key)
            if not isinstance(existing_value, list):
                existing_value = [existing_value]
            existing_value.append(value)
            setattr(obj, last_key, existing_value)
        else:
            setattr(obj, last_key, value)


class QueryTypes(Enum):
    SIMPLE_TABLE = 'SIMPLE_TABLE'
    FUNCTION = 'FUNCTION'
    SEARCH = 'SEARCH'


class MockMicroFetcher:
    def __init__(self, data_mapping=None):
        self.data_mapping = data_mapping or {}
        self.instances = []

    def __call__(self, service_name):
        instance = MockMicroFetcherInstance(service_name, self.data_mapping)
        self.instances.append(instance)
        return instance


class MockMicroFetcherInstance:
    def __init__(self, service_name, data_mapping):
        self.service_name = service_name
        self.data_mapping = data_mapping or {}
        self.app = MagicMock()
        self.base_data = None
        self.query_type = None
        self.output_key = None
        self.fields = []
        self.table_name = None
        self.column_name = None
        self.compair_operator = None
        self.column_values = None
        self.lookup_key = None
        self.module_name = None
        self.function_params = {}
        self.search_list = None
        self.configs = None
        self.global_configs_data = {}
        self.search_configs = {}
        self.function_params = {}
        self.search_configs = {}
        self.search_list = None

    def join(self, base_data, output_key=None):
        self.base_data = base_data
        self.query_type = QueryTypes.SIMPLE_TABLE.value
        if output_key:
            output_key = output_key.split('as ')[1]
        self.output_key = "{}".format(self.service_name.split('_')[0].lower()) if not output_key else output_key
        return self

    def config(self, **configs):
        self.configs = configs
        return self

    def select(self, *args):
        self.fields = list(args)
        return self

    def filter(self, *args):
        against = args[0].split('.')
        self.compair_operator = args[1]
        self.lookup_key = args[2]
        self.column_values = set()
        if isinstance(self.base_data, dict):
            self.column_values.add(get_obj_value(self.base_data, self.lookup_key))
        else:
            data = self.base_data
            if isinstance(data, list):
                for row in data:
                    self.column_values.add(get_obj_value(row, self.lookup_key))
            else:
                self.column_values.add(get_obj_value(data, self.lookup_key))
        self.column_values = list(filter(None, self.column_values))
        if len(self.column_values) == 1:
            self.column_values = self.column_values[0]
        if len(against) != 2:
            self.column_name = against[0]
        else:
            self.table_name = against[0]
            self.column_name = against[1]
            self.fields.append(self.column_name)
        return self

    def fetch(self):
        if self.column_values or self.module_name or self.query_type == QueryTypes.SEARCH.value:
            return self._call_api()
        else:
            return self.base_data

    def execute(self):
        return self.fetch()

    def with_params(self, **kwargs):
        self.function_params = kwargs
        return self

    def from_function(self, module_name):
        self.query_type = QueryTypes.FUNCTION.value
        self.module_name = module_name
        return self

    def global_configs(self, **keywords):
        self.global_configs_data = keywords
        return self

    def feed_list(self, base_data, output_key=None):
        self.join(base_data, output_key)
        self.query_type = QueryTypes.SEARCH.value
        return self

    def search_config(self, configs):
        self.search_configs = configs
        self._prepare_search_list()
        return self

    def _prepare_search_list(self):
        output = dict()
        for index, item in enumerate(self.base_data):
            for search_column in self.search_configs.get("search_priority", []):
                sanitize = None
                if isinstance(search_column, dict):
                    search_column_name = search_column.get('key')
                    operator = search_column.get('operator') or "IN"
                    sanitize = search_column.get('sanitize')
                else:
                    search_column_name = search_column
                    operator = 'IN'

                value = item.get(search_column_name)
                if sanitize and isinstance(sanitize, list):
                    for _san in sanitize:
                        value = _san(value)

                if value:
                    if not output.get(search_column_name):
                        output[search_column_name] = dict(
                            search_key=search_column_name,
                            operator=operator,
                            inputs=dict()
                        )
                    if not output[search_column_name]['inputs'].get(value):
                        output[search_column_name]['inputs'][value] = dict(
                            indexes=[index],
                            search_value=value
                        )
                    else:
                        output[search_column_name]['inputs'][value]["indexes"].append(index)
                    break

        output = list(output.values())
        for item in output:
            item['inputs'] = list(item['inputs'].values())

        self.search_list = output

    def _call_api(self):
        # Use data_mapping to get data to merge

        # Instead of making network calls, we simulate data using data_mapping
        if self.query_type == QueryTypes.SIMPLE_TABLE.value:
            # Simulate data fetching
            # We can use data_mapping to get the data
            key = (self.service_name, self.output_key)
            data_to_merge = self.data_mapping.get(key, [])

            result = data_to_merge  # self.data_mapping.get(self.service_name, [])
        elif self.query_type == QueryTypes.FUNCTION.value:
            key = (self.service_name, self.module_name)
            data_to_merge = self.data_mapping.get(key, [])
            result = data_to_merge
        elif self.query_type == QueryTypes.SEARCH.value:
            # Simulate search
            key = (self.service_name, self.output_key)
            data_to_merge = self.data_mapping.get(key, {})

            if data_to_merge:
                if isinstance(data_to_merge, list):
                    data_to_merge = data_to_merge[0]

                for sl in self.search_list:
                    for inp in sl.get('inputs'):
                        inp.update(dict(matched_id=data_to_merge.get('id')))
                        inp.update(data_to_merge)
            result = self.search_list
        else:
            result = []
        if self.base_data is not None:
            return self._map_base(result)
        return result

    def _map_base(self, result):
        if self.query_type == QueryTypes.SEARCH.value:
            # Map search result with the original object.
            for item in result:
                for _input in item.get('inputs', []):
                    for _index in _input.get('indexes', []):
                        self.base_data[_index][self.output_key] = _input.get('matched_id')
                        append_extra = self.search_configs.get('append_extra') if isinstance(self.search_configs,
                                                                                             dict) else []
                        for _ap_col in append_extra:
                            if _input.get('matched_id'):
                                self.base_data[_index][_ap_col] = _input.get(_ap_col)
                            else:
                                self.base_data[_index][_ap_col] = self.base_data[_index].get(_ap_col)
            validation_result = []
            # for _val in result.get("validation", []):
            #     for _ind in _val.get("indexes", []):
            #         _val.pop("indexes", None)
            #         validation_result.append(dict(
            #             index=_ind,
            #             **_val
            #         ))
            return validation_result
        else:
            if isinstance(self.base_data, dict):
                for rd in result:
                    if self.base_data.get(self.lookup_key) == rd.get(self.column_name):
                        self.base_data[self.output_key] = rd
            else:
                data = self.base_data
                append_if_exists = self.configs.get("append_if_exists", False) if isinstance(self.configs,
                                                                                             dict) else False
                if isinstance(data, list):
                    for row in data:
                        for rd in result:
                            if get_obj_value(row, self.lookup_key) == rd.get(self.column_name):
                                set_obj_value(row, self.output_key, rd, append_if_exists)
                else:
                    for rd in result:
                        if get_obj_value(data, self.lookup_key) == rd.get(self.column_name):
                            set_obj_value(data, self.output_key, rd, append_if_exists)
            return self.base_data



import importlib
import json
import os

from confluent_kafka import Consumer, KafkaError

from wedeliver_core_plus import WedeliverCorePlus
from wedeliver_core_plus.app_decorators import handle_exceptions
from wedeliver_core_plus.helpers.exceptions import AppValidationError


def consume_message(value):
    app = WedeliverCorePlus.get_app()

    message_payload = json.loads(value)

    if not message_payload:
        raise AppValidationError("No message payload found..")

    message = ["Got a message"]
    caller_service = message_payload.get("caller_service")

    if caller_service:
        message.append(f"from service: {caller_service}")

    message.append(f"value: {value}")

    app.logger.info(" ".join(message))

    function_name = message_payload.get("function_name")
    function_params = message_payload.get("function_params") or {}
    if not function_name:
        raise AppValidationError("No function name found..")

    @handle_exceptions
    def _execute_method():
        function_file, function_call = os.path.splitext(function_name)
        module = importlib.import_module(function_file)
        method = getattr(module, function_call[1:])

        return method(**function_params)

    _execute_method()
    # function_result = _execute_method()
    # app.logger.debug(function_result)


def create_consumer(topics):
    """
        Start Kafka listener
        """
    app = WedeliverCorePlus.get_app()

    c = Consumer(
        {
            "bootstrap.servers": app.config.get("BOOTSTRAP_SERVERS"),
            "sasl.username": app.config.get("SASL_USERNAME"),
            "sasl.password": app.config.get("SASL_PASSWORD"),
            "security.protocol": "SASL_SSL",
            "sasl.mechanisms": "PLAIN",
            "group.id": "services_group",
            "auto.offset.reset": "latest",  # earliest or latest
        }
    )

    c.subscribe(topics)

    app.logger.info(f"Kafka Consumer created and subscribed to topics [{' - '.join(topics)}]")

    return c


def poll_messages(consumer, wait_time=3.0):
    app = WedeliverCorePlus.get_app()

    wait_time = wait_time  # timeout is set to 10 second

    app.logger.info(f"Kafka is polling messages for each ({wait_time} seconds)...")

    while True:
        msg = consumer.poll(wait_time)
        # print(1)

        if msg is None:
            continue

        if msg.error():
            if msg.error().code() == KafkaError._PARTITION_EOF:
                continue

            app.logger.error("error: {0}".format(msg.error()))
            break

        # message_key = msg.key()

        value = msg.value().decode('utf-8')

        try:
            consume_message(value)
        except AppValidationError:
            # app.logger.error(traceback.format_exc())
            continue

    consumer.close()



import json
import random
import string
import traceback

from wedeliver_core_plus import WedeliverCorePlus
from confluent_kafka import Producer as kafka_Producer

from wedeliver_core_plus.helpers.exceptions import AppValidationError
from wedeliver_core_plus.helpers.micro_fetcher import MicroFetcher


class Producer:
    app = None

    def __init__(self):
        self.app = WedeliverCorePlus.get_app()
        self._producer = (
            None
            if not self.app or self.app.config.get("FLASK_ENV") == "local"
            else kafka_Producer(
                {
                    "sasl.mechanisms": "PLAIN",
                    "request.timeout.ms": 20000,
                    "bootstrap.servers": self.app.config.get("BOOTSTRAP_SERVERS"),
                    "retry.backoff.ms": 500,
                    "sasl.username": self.app.config.get("SASL_USERNAME"),
                    "sasl.password": self.app.config.get("SASL_PASSWORD"),
                    "security.protocol": "SASL_SSL",
                }
            )
        )

    def send_topic(self, topic, datajson, **configs):

        if not self.app:
            return True

        self.app.logger.debug(datajson)

        micro_fetcher_instead_of_kafka = self.app.config.get("MICRO_FETCHER_INSTEAD_OF_KAFKA", False)

        if micro_fetcher_instead_of_kafka:
            service_name = configs.get('service_name')
            if service_name:
                self.app.logger.debug("Kafka: Direct HTTP Call is enabled")

                MicroFetcher(service_name).from_function(datajson.get('function_name')).with_params(
                    **datajson.get('function_params')).execute()

                return True

            self.app.logger.error(
                f"Kafka ({topic}) No service name found, you need to provide service_name in configs to use MicroFetcher instead of kafka")
            self.app.logger.debug("Kafka: Direct HTTP Call is disabled")


        if self._producer is None:
            self.app.logger.debug(
                "Can not send Kafka message, this is {} environment".format(
                    self.app.config.get("FLASK_ENV")
                )
            )
            return True

        def acked(err, msg):
            """Delivery report handler called on
            successful or failed delivery of message
            """
            if err is not None:

                self.app.logger.error("Failed to deliver message: {0}".format(err))

            else:
                self.app.logger.debug(
                    "Produced record to topic {0} partition [{1}] @ offset {2}".format(
                        msg.topic(), msg.partition(), msg.offset()
                    )
                )
                # is_success = True

        # record_key = "alice"
        record_key = str(random.choices(string.digits, k=5))

        self._producer.produce(
            topic, key=record_key, value=json.dumps(datajson), on_delivery=acked
        )
        self._producer.poll(0)
        self._producer.flush()

        return True



import platform
from datetime import datetime

from wedeliver_core_plus import WedeliverCorePlus
from wedeliver_core_plus.helpers.topics import Topics
from wedeliver_core_plus.helpers import kafka_producers


def send_critical_error(message, channel=None):
    channel = channel or "critical-errors"
    send_notification_message(
        channel=channel,
        title="critical",
        color="#df0000",
        message=message,
        emoji=":pleading_face:"
    )


def send_notification_message(
        message, channel="logs", title="Log", color="#32a4a7", emoji=":dizzy_face:"
):
    app = WedeliverCorePlus.get_app()
    channel = "eng-{0}-{1}".format(
        app.config.get("FLASK_ENV")
        if app.config.get("FLASK_ENV") == "production"
        else "development",
        channel,
    )
    datetime_now = datetime.now().strftime("%Y-%m-%d %H:%M")
    message = "Node: {0}\nEnv:{3}\nDate: {1}\n{2}".format(
        platform.node(), datetime_now, message, str(app.config.get("FLASK_ENV"))
    )
    data = {
        "notification_method": "slack",
        "payload": {
            "channel": channel,
            "title": "New {0} in {1} Service".format(
                title, str(app.config.get("SERVICE_NAME"))
            ),
            "text": message,
            "color": color,
            "icon_emoji": emoji,
        },
    }
    kafka_producers.send_topic(topic=Topics().INTERNAL_NOTIFICATION_MESSAGE, datajson=data)


def send_push_notification(
        message_text, reference_type, user_ids, title=None, payload=None, created_by=None, **kwargs
):
    # token = None

    data = dict(
        user_ids=user_ids,
        reference_type=reference_type,
        created_by=created_by,
        title=title or "weDeliver",
        message_text=message_text,
        payload=payload,
        # language="ar"
    )

    kafka_producers.send_topic(topic=Topics().SEND_PUSH_NOTIFICATION, datajson=dict(
        function_name='app.business_logic.notification.send_push_notification.execute',
        function_params=data
    ), kafka_configs=kwargs.get('kafka_configs'))


def send_slack_notification_message_with_file_content(
        message,file_content, channel="logs", title="Log", color="#32a4a7", emoji=":dizzy_face:"
):
    app = WedeliverCorePlus.get_app()
    channel = "eng-{0}-{1}".format(
        app.config.get("FLASK_ENV")
        if app.config.get("FLASK_ENV") == "production"
        else "development",
        channel,
    )
    datetime_now = datetime.now().strftime("%Y-%m-%d %H:%M")
    message = "Node: {0}\nEnv:{3}\nDate: {1}\n{2}".format(
        platform.node(), datetime_now, message, str(app.config.get("FLASK_ENV"))
    )
    data = {
        "notification_method": "slack_file",
        "payload": {
            "channel": channel,
            "title": "New {0} in {1} Service".format(
                title, str(app.config.get("SERVICE_NAME"))
            ),
            "text": message,
            "color": color,
            "icon_emoji": emoji,
            "file_content":file_content
        },
    }
    kafka_producers.send_topic(topic=Topics().INTERNAL_NOTIFICATION_MESSAGE, datajson=data)



from wedeliver_core_plus.helpers.topics import Topics
from wedeliver_core_plus.helpers import kafka_producers
def send_sms(message, mobile):
    if not isinstance(mobile, list):
        mobile = [mobile]
    kafka_producers.send_topic(topic=Topics().SEND_SMS, datajson=dict(
        function_name='app.business_logic.bulk_sms.send_v2.send_bulk_v2',
        function_params=dict(
            sms_message=message, mobile_numbers=mobile
        )
    ))



def get_obj_value(_obj, key):
    if hasattr(_obj, '__dict__'):  # object
        return _obj.__dict__.get(key)

    elif isinstance(_obj, dict):  # dict
        return _obj.get(key)

    elif hasattr(_obj, key):  # class
        return getattr(_obj, key)


def set_obj_value(_obj, attr_name, attr_value, append_if_exists=False):
    if not append_if_exists:
        try:
            setattr(_obj, attr_name, attr_value)
        except AttributeError:
            _obj[attr_name] = attr_value
    else:
        try:
            old_value = getattr(_obj, attr_name) if hasattr(_obj, attr_name) else None
            if old_value:
                if isinstance(old_value, dict):
                    setattr(_obj, attr_name, [old_value])
                getattr(_obj, attr_name).append(attr_value)
            else:
                setattr(_obj, attr_name, attr_value)
        except AttributeError:
            old_value = _obj.get(attr_name)
            if old_value:
                if isinstance(old_value, dict):
                    _obj[attr_name] = [old_value]

                _obj[attr_name].append(attr_value)
            else:
                _obj[attr_name] = attr_value
